#! /usr/bin/env python

import random
import sys
import math


from pandac.PandaModules import loadPrcFile
loadPrcFile("settings.prc")

import direct.directbase.DirectStart
from direct.showbase import DirectObject
from pandac.PandaModules import *


# This block of text exists to provide a basic explanation of what is going on.
# The first thing to realise is that these are not portals in the walk in at one location, walk out at another location sense - they only exist to optimise rendering speed by culling large amounts of geometry.
# Nothing changes location, or is perceived to have changed location.
# The idea is that you divide the level into cells, where each cell has limited visibility of the other cells.
# Each cell is represented by a PandaNode, with everything in that cell a child of the node.
# You then have portals within each cell - a portal is a single polygon with 4 vertices through which you can see into another cell.
# For instance two cells might be a room and a coridoor, so you would have a portal in the doorway.
# Portals are one way, going from the cell they are in to another cell; this means they will often come in pairs, one for each way.
# What they do is when you can see a portal it checks and renders only stuff that can be seen from the cameras viewpoint, that is in the 'out' cell.
# This works recursivly, so if you can see a portal through a portal then that portal will also go through its children and check visibility, and so on recursivly.
# Whilst this is not much use in outdoor environments in indoor environments this method could easilly save rendering the vast majority of a level.
# The below code demonstrates usage but a key issue is that all cells must be hidden except the one that the camera is in.
# The currentCellTask task acheives this below.
# Additionally, if an object were to move from one cell to the next its parent cell would need to be updated - the PortalNode only provides basic clipping functionality, the rest has to be done in code.
# There is also a potential visibility glitch in that a portal could be not visible but an object inside its 'out' cell could be poking through and visible, but because of the portal code not rendered - this can be ignored or has to be solved by the user. (Its usually rare that a user will see it, and only happens with very particular level geometry/object combos.)
# As a further point you need to enable portal culling in the settings file - this is done by the settings file provided with this example.


# Create the cells...
cell1 = render.attachNewNode('cell1')
cell2 = render.attachNewNode('cell2')
cell3 = render.attachNewNode('cell3')

cell1.setPos(-10.0,0.0,0.0)
cell2.setPos(10.0,0.0,0.0)
cell3.setPos(30.0,0.0,0.0)


# Add some objects...
for  i in xrange(1,75):
  box = loader.loadModel('simplebox')
  parent = random.choice([cell1,cell2,cell3])
  box.reparentTo(parent)
  box.setPos(random.uniform(-5.0,5.0),random.uniform(-5.0,5.0),random.uniform(-5.0,5.0))


# Create objects to represent the edges of the portals - we actually make two sets and arrange a button press to flip between the two...
edges = []

edgeA = loader.loadModel('portal_edge')
edgeA.reparentTo(render)
edges.append(edgeA)

edgeA2 = loader.loadModel('portal_edge')
edgeA2.setPos(20.0,0.0,0.0)
edgeA2.reparentTo(render)
edges.append(edgeA2)

edgeB = loader.loadModel('portal_edge2')
edgeB.reparentTo(render)
edgeB.hide()
edges.append(edgeB)

edgeB2 = loader.loadModel('portal_edge2')
edgeB2.setPos(20.0,0.0,0.0)
edgeB2.reparentTo(render)
edgeB2.hide()
edges.append(edgeB2)


class PortalSwitch(DirectObject.DirectObject):
  def __init__(self):
    self.accept('p',self.switch)
    
  def switch(self):
    for edge in edges:
      if edge.isHidden():
        edge.show()
      else:
        edge.hide()
        
portalSwitch = PortalSwitch()


# Create the portal going from cell1 to cell2...
from1 = PortalNode('1to2')
from1Node = cell1.attachNewNode(from1)
from1Node.setPos(render,Vec3(0.0,0.0,0.0))

from1.clearVertices()
from1.addVertex(Point3(0.0,3.0,3.0))
from1.addVertex(Point3(0.0,3.0,-3.0))
from1.addVertex(Point3(0.0,-3.0,-3.0))
from1.addVertex(Point3(0.0,-3.0,3.0))

from1.setCellIn(cell1)
from1.setCellOut(cell2)


# Create the portal going from cell2 to cell1...
from2 = PortalNode('2to1')
from2Node = cell2.attachNewNode(from2)
from2Node.setPos(render,Vec3(0.0,0.0,0.0))

from2.clearVertices()
from2.addVertex(Point3(0.0,3.0,3.0))
from2.addVertex(Point3(0.0,-3.0,3.0))
from2.addVertex(Point3(0.0,-3.0,-3.0))
from2.addVertex(Point3(0.0,3.0,-3.0))

from2.setCellIn(cell2)
from2.setCellOut(cell1)


# Create the portal going from cell2 to cell3 (Don't bother with 3 to 2 as the camera never goes there.)...
from3 = PortalNode('2to3')
from3Node = cell1.attachNewNode(from3)
from3Node.setPos(render,Vec3(20.0,0.0,0.0))

from3.clearVertices()
from3.addVertex(Point3(0.0,3.0,3.0))
from3.addVertex(Point3(0.0,3.0,-3.0))
from3.addVertex(Point3(0.0,-3.0,-3.0))
from3.addVertex(Point3(0.0,-3.0,3.0))

from3.setCellIn(cell2)
from3.setCellOut(cell3)


# Create a task that detects which cell the camera is in and makes teh relevant cell visible, whilst making all others hidden...
def currentCellTask(task):
  if base.camera.getPos()[0]<0.0:
    cell1.show()
    cell2.hide()
  else:
    cell1.hide()
    cell2.show()
  cell3.hide()
  return task.cont

taskMgr.add(currentCellTask,'currentCellTask')


# Add some light...
light = PointLight('light')
lightNode = base.camera.attachNewNode(light)
render.setLight(lightNode)


# Arrange for the camera to fly arround suitably...
base.disableMouse()

def camTask(task):
  time = globalClock.getRealTime()
  ang = (math.fmod(time,8.0)/8.0)*math.pi*2.0

  base.camera.setPos(19.0*math.cos(ang),19.0*math.sin(ang),0.0)
  base.camera.lookAt(0.0,0.0,0.0)
  return task.cont

taskMgr.add(camTask,'camTask')


# Escape to exit...
class Killer(DirectObject.DirectObject):
  def __init__(self):
    self.accept('escape',sys.exit)
killer = Killer()


# Go...
run()